#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: org/apache/spark/sql/execution/streaming/StateMessage.proto
# Protobuf Python Version: 6.33.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    33,
    0,
    "",
    "org/apache/spark/sql/execution/streaming/StateMessage.proto",
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n;org/apache/spark/sql/execution/streaming/StateMessage.proto\x12.org.apache.spark.sql.execution.streaming.state"\x84\x05\n\x0cStateRequest\x12\x18\n\x07version\x18\x01 \x01(\x05R\x07version\x12}\n\x15statefulProcessorCall\x18\x02 \x01(\x0b\x32\x45.org.apache.spark.sql.execution.streaming.state.StatefulProcessorCallH\x00R\x15statefulProcessorCall\x12z\n\x14stateVariableRequest\x18\x03 \x01(\x0b\x32\x44.org.apache.spark.sql.execution.streaming.state.StateVariableRequestH\x00R\x14stateVariableRequest\x12\x8c\x01\n\x1aimplicitGroupingKeyRequest\x18\x04 \x01(\x0b\x32J.org.apache.spark.sql.execution.streaming.state.ImplicitGroupingKeyRequestH\x00R\x1aimplicitGroupingKeyRequest\x12\x62\n\x0ctimerRequest\x18\x05 \x01(\x0b\x32<.org.apache.spark.sql.execution.streaming.state.TimerRequestH\x00R\x0ctimerRequest\x12\x62\n\x0cutilsRequest\x18\x06 \x01(\x0b\x32<.org.apache.spark.sql.execution.streaming.state.UtilsRequestH\x00R\x0cutilsRequestB\x08\n\x06method"i\n\rStateResponse\x12\x1e\n\nstatusCode\x18\x01 \x01(\x05R\nstatusCode\x12"\n\x0c\x65rrorMessage\x18\x02 \x01(\tR\x0c\x65rrorMessage\x12\x14\n\x05value\x18\x03 \x01(\x0cR\x05value"x\n\x1cStateResponseWithLongTypeVal\x12\x1e\n\nstatusCode\x18\x01 \x01(\x05R\nstatusCode\x12"\n\x0c\x65rrorMessage\x18\x02 \x01(\tR\x0c\x65rrorMessage\x12\x14\n\x05value\x18\x03 \x01(\x03R\x05value"z\n\x1eStateResponseWithStringTypeVal\x12\x1e\n\nstatusCode\x18\x01 \x01(\x05R\nstatusCode\x12"\n\x0c\x65rrorMessage\x18\x02 \x01(\tR\x0c\x65rrorMessage\x12\x14\n\x05value\x18\x03 \x01(\tR\x05value"\xa0\x01\n\x18StateResponseWithListGet\x12\x1e\n\nstatusCode\x18\x01 \x01(\x05R\nstatusCode\x12"\n\x0c\x65rrorMessage\x18\x02 \x01(\tR\x0c\x65rrorMessage\x12\x14\n\x05value\x18\x03 \x03(\x0cR\x05value\x12*\n\x10requireNextFetch\x18\x04 \x01(\x08R\x10requireNextFetch"\xa8\x01\n StateResponseWithMapKeysOrValues\x12\x1e\n\nstatusCode\x18\x01 \x01(\x05R\nstatusCode\x12"\n\x0c\x65rrorMessage\x18\x02 \x01(\tR\x0c\x65rrorMessage\x12\x14\n\x05value\x18\x03 \x03(\x0cR\x05value\x12*\n\x10requireNextFetch\x18\x04 \x01(\x08R\x10requireNextFetch"9\n\x0fKeyAndValuePair\x12\x10\n\x03key\x18\x01 \x01(\x0cR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x0cR\x05value"\xe7\x01\n\x1cStateResponseWithMapIterator\x12\x1e\n\nstatusCode\x18\x01 \x01(\x05R\nstatusCode\x12"\n\x0c\x65rrorMessage\x18\x02 \x01(\tR\x0c\x65rrorMessage\x12W\n\x06kvPair\x18\x03 \x03(\x0b\x32?.org.apache.spark.sql.execution.streaming.state.KeyAndValuePairR\x06kvPair\x12*\n\x10requireNextFetch\x18\x04 \x01(\x08R\x10requireNextFetch"L\n\tTimerInfo\x12\x15\n\x03key\x18\x01 \x01(\x0cH\x00R\x03key\x88\x01\x01\x12 \n\x0btimestampMs\x18\x02 \x01(\x03R\x0btimestampMsB\x06\n\x04_key"\xd9\x01\n\x16StateResponseWithTimer\x12\x1e\n\nstatusCode\x18\x01 \x01(\x05R\nstatusCode\x12"\n\x0c\x65rrorMessage\x18\x02 \x01(\tR\x0c\x65rrorMessage\x12O\n\x05timer\x18\x03 \x03(\x0b\x32\x39.org.apache.spark.sql.execution.streaming.state.TimerInfoR\x05timer\x12*\n\x10requireNextFetch\x18\x04 \x01(\x08R\x10requireNextFetch"\xa0\x05\n\x15StatefulProcessorCall\x12h\n\x0esetHandleState\x18\x01 \x01(\x0b\x32>.org.apache.spark.sql.execution.streaming.state.SetHandleStateH\x00R\x0esetHandleState\x12h\n\rgetValueState\x18\x02 \x01(\x0b\x32@.org.apache.spark.sql.execution.streaming.state.StateCallCommandH\x00R\rgetValueState\x12\x66\n\x0cgetListState\x18\x03 \x01(\x0b\x32@.org.apache.spark.sql.execution.streaming.state.StateCallCommandH\x00R\x0cgetListState\x12\x64\n\x0bgetMapState\x18\x04 \x01(\x0b\x32@.org.apache.spark.sql.execution.streaming.state.StateCallCommandH\x00R\x0bgetMapState\x12o\n\x0etimerStateCall\x18\x05 \x01(\x0b\x32\x45.org.apache.spark.sql.execution.streaming.state.TimerStateCallCommandH\x00R\x0etimerStateCall\x12j\n\x0e\x64\x65leteIfExists\x18\x06 \x01(\x0b\x32@.org.apache.spark.sql.execution.streaming.state.StateCallCommandH\x00R\x0e\x64\x65leteIfExistsB\x08\n\x06method"\xd5\x02\n\x14StateVariableRequest\x12h\n\x0evalueStateCall\x18\x01 \x01(\x0b\x32>.org.apache.spark.sql.execution.streaming.state.ValueStateCallH\x00R\x0evalueStateCall\x12\x65\n\rlistStateCall\x18\x02 \x01(\x0b\x32=.org.apache.spark.sql.execution.streaming.state.ListStateCallH\x00R\rlistStateCall\x12\x62\n\x0cmapStateCall\x18\x03 \x01(\x0b\x32<.org.apache.spark.sql.execution.streaming.state.MapStateCallH\x00R\x0cmapStateCallB\x08\n\x06method"\x83\x02\n\x1aImplicitGroupingKeyRequest\x12h\n\x0esetImplicitKey\x18\x01 \x01(\x0b\x32>.org.apache.spark.sql.execution.streaming.state.SetImplicitKeyH\x00R\x0esetImplicitKey\x12q\n\x11removeImplicitKey\x18\x02 \x01(\x0b\x32\x41.org.apache.spark.sql.execution.streaming.state.RemoveImplicitKeyH\x00R\x11removeImplicitKeyB\x08\n\x06method"\x81\x02\n\x0cTimerRequest\x12q\n\x11timerValueRequest\x18\x01 \x01(\x0b\x32\x41.org.apache.spark.sql.execution.streaming.state.TimerValueRequestH\x00R\x11timerValueRequest\x12t\n\x12\x65xpiryTimerRequest\x18\x02 \x01(\x0b\x32\x42.org.apache.spark.sql.execution.streaming.state.ExpiryTimerRequestH\x00R\x12\x65xpiryTimerRequestB\x08\n\x06method"\xf6\x01\n\x11TimerValueRequest\x12s\n\x12getProcessingTimer\x18\x01 \x01(\x0b\x32\x41.org.apache.spark.sql.execution.streaming.state.GetProcessingTimeH\x00R\x12getProcessingTimer\x12\x62\n\x0cgetWatermark\x18\x02 \x01(\x0b\x32<.org.apache.spark.sql.execution.streaming.state.GetWatermarkH\x00R\x0cgetWatermarkB\x08\n\x06method"b\n\x12\x45xpiryTimerRequest\x12\x1e\n\niteratorId\x18\x01 \x01(\tR\niteratorId\x12,\n\x11\x65xpiryTimestampMs\x18\x02 \x01(\x03R\x11\x65xpiryTimestampMs"\x13\n\x11GetProcessingTime"\x0e\n\x0cGetWatermark"\x8b\x01\n\x0cUtilsRequest\x12q\n\x11parseStringSchema\x18\x01 \x01(\x0b\x32\x41.org.apache.spark.sql.execution.streaming.state.ParseStringSchemaH\x00R\x11parseStringSchemaB\x08\n\x06method"+\n\x11ParseStringSchema\x12\x16\n\x06schema\x18\x01 \x01(\tR\x06schema"\xc7\x01\n\x10StateCallCommand\x12\x1c\n\tstateName\x18\x01 \x01(\tR\tstateName\x12\x16\n\x06schema\x18\x02 \x01(\tR\x06schema\x12\x30\n\x13mapStateValueSchema\x18\x03 \x01(\tR\x13mapStateValueSchema\x12K\n\x03ttl\x18\x04 \x01(\x0b\x32\x39.org.apache.spark.sql.execution.streaming.state.TTLConfigR\x03ttl"\xa7\x02\n\x15TimerStateCallCommand\x12[\n\x08register\x18\x01 \x01(\x0b\x32=.org.apache.spark.sql.execution.streaming.state.RegisterTimerH\x00R\x08register\x12U\n\x06\x64\x65lete\x18\x02 \x01(\x0b\x32;.org.apache.spark.sql.execution.streaming.state.DeleteTimerH\x00R\x06\x64\x65lete\x12P\n\x04list\x18\x03 \x01(\x0b\x32:.org.apache.spark.sql.execution.streaming.state.ListTimersH\x00R\x04listB\x08\n\x06method"\x92\x03\n\x0eValueStateCall\x12\x1c\n\tstateName\x18\x01 \x01(\tR\tstateName\x12P\n\x06\x65xists\x18\x02 \x01(\x0b\x32\x36.org.apache.spark.sql.execution.streaming.state.ExistsH\x00R\x06\x65xists\x12G\n\x03get\x18\x03 \x01(\x0b\x32\x33.org.apache.spark.sql.execution.streaming.state.GetH\x00R\x03get\x12n\n\x10valueStateUpdate\x18\x04 \x01(\x0b\x32@.org.apache.spark.sql.execution.streaming.state.ValueStateUpdateH\x00R\x10valueStateUpdate\x12M\n\x05\x63lear\x18\x05 \x01(\x0b\x32\x35.org.apache.spark.sql.execution.streaming.state.ClearH\x00R\x05\x63learB\x08\n\x06method"\xdf\x04\n\rListStateCall\x12\x1c\n\tstateName\x18\x01 \x01(\tR\tstateName\x12P\n\x06\x65xists\x18\x02 \x01(\x0b\x32\x36.org.apache.spark.sql.execution.streaming.state.ExistsH\x00R\x06\x65xists\x12\x62\n\x0clistStateGet\x18\x03 \x01(\x0b\x32<.org.apache.spark.sql.execution.streaming.state.ListStateGetH\x00R\x0clistStateGet\x12\x62\n\x0clistStatePut\x18\x04 \x01(\x0b\x32<.org.apache.spark.sql.execution.streaming.state.ListStatePutH\x00R\x0clistStatePut\x12_\n\x0b\x61ppendValue\x18\x05 \x01(\x0b\x32;.org.apache.spark.sql.execution.streaming.state.AppendValueH\x00R\x0b\x61ppendValue\x12\\\n\nappendList\x18\x06 \x01(\x0b\x32:.org.apache.spark.sql.execution.streaming.state.AppendListH\x00R\nappendList\x12M\n\x05\x63lear\x18\x07 \x01(\x0b\x32\x35.org.apache.spark.sql.execution.streaming.state.ClearH\x00R\x05\x63learB\x08\n\x06method"\xc2\x06\n\x0cMapStateCall\x12\x1c\n\tstateName\x18\x01 \x01(\tR\tstateName\x12P\n\x06\x65xists\x18\x02 \x01(\x0b\x32\x36.org.apache.spark.sql.execution.streaming.state.ExistsH\x00R\x06\x65xists\x12V\n\x08getValue\x18\x03 \x01(\x0b\x32\x38.org.apache.spark.sql.execution.streaming.state.GetValueH\x00R\x08getValue\x12_\n\x0b\x63ontainsKey\x18\x04 \x01(\x0b\x32;.org.apache.spark.sql.execution.streaming.state.ContainsKeyH\x00R\x0b\x63ontainsKey\x12_\n\x0bupdateValue\x18\x05 \x01(\x0b\x32;.org.apache.spark.sql.execution.streaming.state.UpdateValueH\x00R\x0bupdateValue\x12V\n\x08iterator\x18\x06 \x01(\x0b\x32\x38.org.apache.spark.sql.execution.streaming.state.IteratorH\x00R\x08iterator\x12J\n\x04keys\x18\x07 \x01(\x0b\x32\x34.org.apache.spark.sql.execution.streaming.state.KeysH\x00R\x04keys\x12P\n\x06values\x18\x08 \x01(\x0b\x32\x36.org.apache.spark.sql.execution.streaming.state.ValuesH\x00R\x06values\x12Y\n\tremoveKey\x18\t \x01(\x0b\x32\x39.org.apache.spark.sql.execution.streaming.state.RemoveKeyH\x00R\tremoveKey\x12M\n\x05\x63lear\x18\n \x01(\x0b\x32\x35.org.apache.spark.sql.execution.streaming.state.ClearH\x00R\x05\x63learB\x08\n\x06method""\n\x0eSetImplicitKey\x12\x10\n\x03key\x18\x01 \x01(\x0cR\x03key"\x13\n\x11RemoveImplicitKey"\x08\n\x06\x45xists"\x05\n\x03Get"=\n\rRegisterTimer\x12,\n\x11\x65xpiryTimestampMs\x18\x01 \x01(\x03R\x11\x65xpiryTimestampMs";\n\x0b\x44\x65leteTimer\x12,\n\x11\x65xpiryTimestampMs\x18\x01 \x01(\x03R\x11\x65xpiryTimestampMs",\n\nListTimers\x12\x1e\n\niteratorId\x18\x01 \x01(\tR\niteratorId"(\n\x10ValueStateUpdate\x12\x14\n\x05value\x18\x01 \x01(\x0cR\x05value"\x07\n\x05\x43lear".\n\x0cListStateGet\x12\x1e\n\niteratorId\x18\x01 \x01(\tR\niteratorId"L\n\x0cListStatePut\x12\x14\n\x05value\x18\x01 \x03(\x0cR\x05value\x12&\n\x0e\x66\x65tchWithArrow\x18\x02 \x01(\x08R\x0e\x66\x65tchWithArrow"#\n\x0b\x41ppendValue\x12\x14\n\x05value\x18\x01 \x01(\x0cR\x05value"J\n\nAppendList\x12\x14\n\x05value\x18\x01 \x03(\x0cR\x05value\x12&\n\x0e\x66\x65tchWithArrow\x18\x02 \x01(\x08R\x0e\x66\x65tchWithArrow"$\n\x08GetValue\x12\x18\n\x07userKey\x18\x01 \x01(\x0cR\x07userKey"\'\n\x0b\x43ontainsKey\x12\x18\n\x07userKey\x18\x01 \x01(\x0cR\x07userKey"=\n\x0bUpdateValue\x12\x18\n\x07userKey\x18\x01 \x01(\x0cR\x07userKey\x12\x14\n\x05value\x18\x02 \x01(\x0cR\x05value"*\n\x08Iterator\x12\x1e\n\niteratorId\x18\x01 \x01(\tR\niteratorId"&\n\x04Keys\x12\x1e\n\niteratorId\x18\x01 \x01(\tR\niteratorId"(\n\x06Values\x12\x1e\n\niteratorId\x18\x01 \x01(\tR\niteratorId"%\n\tRemoveKey\x12\x18\n\x07userKey\x18\x01 \x01(\x0cR\x07userKey"c\n\x0eSetHandleState\x12Q\n\x05state\x18\x01 \x01(\x0e\x32;.org.apache.spark.sql.execution.streaming.state.HandleStateR\x05state"+\n\tTTLConfig\x12\x1e\n\ndurationMs\x18\x01 \x01(\x03R\ndurationMs*n\n\x0bHandleState\x12\x0c\n\x08PRE_INIT\x10\x00\x12\x0b\n\x07\x43REATED\x10\x01\x12\x0f\n\x0bINITIALIZED\x10\x02\x12\x12\n\x0e\x44\x41TA_PROCESSED\x10\x03\x12\x13\n\x0fTIMER_PROCESSED\x10\x04\x12\n\n\x06\x43LOSED\x10\x05\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "pyspark.sql.streaming.proto.StateMessage_pb2", _globals
)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_HANDLESTATE"]._serialized_start = 7489
    _globals["_HANDLESTATE"]._serialized_end = 7599
    _globals["_STATEREQUEST"]._serialized_start = 112
    _globals["_STATEREQUEST"]._serialized_end = 756
    _globals["_STATERESPONSE"]._serialized_start = 758
    _globals["_STATERESPONSE"]._serialized_end = 863
    _globals["_STATERESPONSEWITHLONGTYPEVAL"]._serialized_start = 865
    _globals["_STATERESPONSEWITHLONGTYPEVAL"]._serialized_end = 985
    _globals["_STATERESPONSEWITHSTRINGTYPEVAL"]._serialized_start = 987
    _globals["_STATERESPONSEWITHSTRINGTYPEVAL"]._serialized_end = 1109
    _globals["_STATERESPONSEWITHLISTGET"]._serialized_start = 1112
    _globals["_STATERESPONSEWITHLISTGET"]._serialized_end = 1272
    _globals["_STATERESPONSEWITHMAPKEYSORVALUES"]._serialized_start = 1275
    _globals["_STATERESPONSEWITHMAPKEYSORVALUES"]._serialized_end = 1443
    _globals["_KEYANDVALUEPAIR"]._serialized_start = 1445
    _globals["_KEYANDVALUEPAIR"]._serialized_end = 1502
    _globals["_STATERESPONSEWITHMAPITERATOR"]._serialized_start = 1505
    _globals["_STATERESPONSEWITHMAPITERATOR"]._serialized_end = 1736
    _globals["_TIMERINFO"]._serialized_start = 1738
    _globals["_TIMERINFO"]._serialized_end = 1814
    _globals["_STATERESPONSEWITHTIMER"]._serialized_start = 1817
    _globals["_STATERESPONSEWITHTIMER"]._serialized_end = 2034
    _globals["_STATEFULPROCESSORCALL"]._serialized_start = 2037
    _globals["_STATEFULPROCESSORCALL"]._serialized_end = 2709
    _globals["_STATEVARIABLEREQUEST"]._serialized_start = 2712
    _globals["_STATEVARIABLEREQUEST"]._serialized_end = 3053
    _globals["_IMPLICITGROUPINGKEYREQUEST"]._serialized_start = 3056
    _globals["_IMPLICITGROUPINGKEYREQUEST"]._serialized_end = 3315
    _globals["_TIMERREQUEST"]._serialized_start = 3318
    _globals["_TIMERREQUEST"]._serialized_end = 3575
    _globals["_TIMERVALUEREQUEST"]._serialized_start = 3578
    _globals["_TIMERVALUEREQUEST"]._serialized_end = 3824
    _globals["_EXPIRYTIMERREQUEST"]._serialized_start = 3826
    _globals["_EXPIRYTIMERREQUEST"]._serialized_end = 3924
    _globals["_GETPROCESSINGTIME"]._serialized_start = 3926
    _globals["_GETPROCESSINGTIME"]._serialized_end = 3945
    _globals["_GETWATERMARK"]._serialized_start = 3947
    _globals["_GETWATERMARK"]._serialized_end = 3961
    _globals["_UTILSREQUEST"]._serialized_start = 3964
    _globals["_UTILSREQUEST"]._serialized_end = 4103
    _globals["_PARSESTRINGSCHEMA"]._serialized_start = 4105
    _globals["_PARSESTRINGSCHEMA"]._serialized_end = 4148
    _globals["_STATECALLCOMMAND"]._serialized_start = 4151
    _globals["_STATECALLCOMMAND"]._serialized_end = 4350
    _globals["_TIMERSTATECALLCOMMAND"]._serialized_start = 4353
    _globals["_TIMERSTATECALLCOMMAND"]._serialized_end = 4648
    _globals["_VALUESTATECALL"]._serialized_start = 4651
    _globals["_VALUESTATECALL"]._serialized_end = 5053
    _globals["_LISTSTATECALL"]._serialized_start = 5056
    _globals["_LISTSTATECALL"]._serialized_end = 5663
    _globals["_MAPSTATECALL"]._serialized_start = 5666
    _globals["_MAPSTATECALL"]._serialized_end = 6500
    _globals["_SETIMPLICITKEY"]._serialized_start = 6502
    _globals["_SETIMPLICITKEY"]._serialized_end = 6536
    _globals["_REMOVEIMPLICITKEY"]._serialized_start = 6538
    _globals["_REMOVEIMPLICITKEY"]._serialized_end = 6557
    _globals["_EXISTS"]._serialized_start = 6559
    _globals["_EXISTS"]._serialized_end = 6567
    _globals["_GET"]._serialized_start = 6569
    _globals["_GET"]._serialized_end = 6574
    _globals["_REGISTERTIMER"]._serialized_start = 6576
    _globals["_REGISTERTIMER"]._serialized_end = 6637
    _globals["_DELETETIMER"]._serialized_start = 6639
    _globals["_DELETETIMER"]._serialized_end = 6698
    _globals["_LISTTIMERS"]._serialized_start = 6700
    _globals["_LISTTIMERS"]._serialized_end = 6744
    _globals["_VALUESTATEUPDATE"]._serialized_start = 6746
    _globals["_VALUESTATEUPDATE"]._serialized_end = 6786
    _globals["_CLEAR"]._serialized_start = 6788
    _globals["_CLEAR"]._serialized_end = 6795
    _globals["_LISTSTATEGET"]._serialized_start = 6797
    _globals["_LISTSTATEGET"]._serialized_end = 6843
    _globals["_LISTSTATEPUT"]._serialized_start = 6845
    _globals["_LISTSTATEPUT"]._serialized_end = 6921
    _globals["_APPENDVALUE"]._serialized_start = 6923
    _globals["_APPENDVALUE"]._serialized_end = 6958
    _globals["_APPENDLIST"]._serialized_start = 6960
    _globals["_APPENDLIST"]._serialized_end = 7034
    _globals["_GETVALUE"]._serialized_start = 7036
    _globals["_GETVALUE"]._serialized_end = 7072
    _globals["_CONTAINSKEY"]._serialized_start = 7074
    _globals["_CONTAINSKEY"]._serialized_end = 7113
    _globals["_UPDATEVALUE"]._serialized_start = 7115
    _globals["_UPDATEVALUE"]._serialized_end = 7176
    _globals["_ITERATOR"]._serialized_start = 7178
    _globals["_ITERATOR"]._serialized_end = 7220
    _globals["_KEYS"]._serialized_start = 7222
    _globals["_KEYS"]._serialized_end = 7260
    _globals["_VALUES"]._serialized_start = 7262
    _globals["_VALUES"]._serialized_end = 7302
    _globals["_REMOVEKEY"]._serialized_start = 7304
    _globals["_REMOVEKEY"]._serialized_end = 7341
    _globals["_SETHANDLESTATE"]._serialized_start = 7343
    _globals["_SETHANDLESTATE"]._serialized_end = 7442
    _globals["_TTLCONFIG"]._serialized_start = 7444
    _globals["_TTLCONFIG"]._serialized_end = 7487
# @@protoc_insertion_point(module_scope)
